from django.urls import path

from . import views

app_name = "blueprints"

urlpatterns = [
    path("", views.index, name="index"),
    path(
        "modals/view_blueprint", views.view_blueprint_modal, name="view_blueprint_modal"
    ),
    path("modals/view_request", views.view_request_modal, name="view_request_modal"),
    path(
        "owners/add/character",
        views.add_personal_blueprint_owner,
        name="add_personal_blueprint_owner",
    ),
    path(
        "owners/add/corporation",
        views.add_corporate_blueprint_owner,
        name="add_corporate_blueprint_owner",
    ),
    path(
        "owners/user",
        views.list_user_owners,
        name="list_user_owners",
    ),
    path(
        "owners/<int:owner_id>/remove",
        views.remove_owner,
        name="remove_owner",
    ),
    path("blueprints", views.list_blueprints, name="list_blueprints"),
    path("requests/user", views.list_user_requests, name="list_user_requests"),
    path("requests/open", views.list_open_requests, name="list_open_requests"),
    path("requests/add", views.create_request, name="create_request"),
    path(
        "requests/<int:request_id>/open", views.mark_request_open, name="request_open"
    ),
    path(
        "requests/<int:request_id>/fulfill",
        views.mark_request_fulfilled,
        name="request_fulfilled",
    ),
    path(
        "requests/<int:request_id>/in_progress",
        views.mark_request_in_progress,
        name="request_in_progress",
    ),
    path(
        "requests/<int:request_id>/cancel",
        views.mark_request_cancelled,
        name="request_cancelled",
    ),
]



# Eve IDs
EVE_CATEGORY_ID_STATION = 3
EVE_CATEGORY_ID_SHIP = 6
EVE_CATEGORY_ID_MODULE = 7
EVE_CATEGORY_ID_CHARGE = 8
EVE_CATEGORY_ID_BLUEPRINT = 9
EVE_CATEGORY_ID_SKILL = 16
EVE_CATEGORY_ID_DRONE = 18
EVE_CATEGORY_ID_IMPLANT = 20
EVE_CATEGORY_ID_FIGHTER = 87
EVE_CATEGORY_ID_STRUCTURE = 65
EVE_TYPE_ID_SOLAR_SYSTEM = 5

EVE_LOCATION_FLAGS = [
    "AssetSafety",
    "AutoFit",
    "Bonus",
    "Booster",
    "BoosterBay",
    "Capsule",
    "Cargo",
    "CorpDeliveries",
    "CorpSAG1",
    "CorpSAG2",
    "CorpSAG3",
    "CorpSAG4",
    "CorpSAG5",
    "CorpSAG6",
    "CorpSAG7",
    "CrateLoot",
    "Deliveries",
    "DroneBay",
    "DustBattle",
    "DustDatabank",
    "FighterBay",
    "FighterTube0",
    "FighterTube1",
    "FighterTube2",
    "FighterTube3",
    "FighterTube4",
    "FleetHangar",
    "FrigateEscapeBay",
    "Hangar",
    "HangarAll",
    "HiSlot0",
    "HiSlot1",
    "HiSlot2",
    "HiSlot3",
    "HiSlot4",
    "HiSlot5",
    "HiSlot6",
    "HiSlot7",
    "HiddenModifiers",
    "Implant",
    "Impounded",
    "JunkyardReprocessed",
    "JunkyardTrashed",
    "LoSlot0",
    "LoSlot1",
    "LoSlot2",
    "LoSlot3",
    "LoSlot4",
    "LoSlot5",
    "LoSlot6",
    "LoSlot7",
    "Locked",
    "MedSlot0",
    "MedSlot1",
    "MedSlot2",
    "MedSlot3",
    "MedSlot4",
    "MedSlot5",
    "MedSlot6",
    "MedSlot7",
    "OfficeFolder",
    "Pilot",
    "PlanetSurface",
    "QuafeBay",
    "QuantumCoreRoom",
    "Reward",
    "RigSlot0",
    "RigSlot1",
    "RigSlot2",
    "RigSlot3",
    "RigSlot4",
    "RigSlot5",
    "RigSlot6",
    "RigSlot7",
    "SecondaryStorage",
    "ServiceSlot0",
    "ServiceSlot1",
    "ServiceSlot2",
    "ServiceSlot3",
    "ServiceSlot4",
    "ServiceSlot5",
    "ServiceSlot6",
    "ServiceSlot7",
    "ShipHangar",
    "ShipOffline",
    "Skill",
    "SkillInTraining",
    "SpecializedAmmoHold",
    "SpecializedCommandCenterHold",
    "SpecializedFuelBay",
    "SpecializedGasHold",
    "SpecializedIndustrialShipHold",
    "SpecializedLargeShipHold",
    "SpecializedMaterialBay",
    "SpecializedMediumShipHold",
    "SpecializedMineralHold",
    "SpecializedOreHold",
    "SpecializedPlanetaryCommoditiesHold",
    "SpecializedSalvageHold",
    "SpecializedShipHold",
    "SpecializedSmallShipHold",
    "StructureActive",
    "StructureFuel",
    "StructureInactive",
    "StructureOffline",
    "SubSystemBay",
    "SubSystemSlot0",
    "SubSystemSlot1",
    "SubSystemSlot2",
    "SubSystemSlot3",
    "SubSystemSlot4",
    "SubSystemSlot5",
    "SubSystemSlot6",
    "SubSystemSlot7",
    "Unlocked",
    "Wallet",
    "Wardrobe",
]



from typing import Tuple

from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import gettext_lazy as _

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.evelinks import dotlan
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveEntity, EveSolarSystem, EveType

from . import __title__
from .constants import EVE_LOCATION_FLAGS
from .decorators import fetch_token_for_owner
from .managers import LocationManager, RequestManager
from .providers import esi
from .utils import LoggerAddTag, make_logger_prefix
from .validators import validate_material_efficiency, validate_time_efficiency

NAMES_MAX_LENGTH = 100

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app"),
            ("request_blueprints", "Can request blueprints"),
            ("manage_requests", "Can review and accept blueprint requests"),
            ("add_personal_blueprint_owner", "Can add personal blueprint owners"),
            ("add_corporate_blueprint_owner", "Can add corporate blueprint owners"),
            ("view_alliance_blueprints", "Can view alliance's blueprints"),
            ("view_industry_jobs", "Can view details about running industry jobs"),
        )


class Owner(models.Model):
    """A corporation that owns blueprints"""

    ERROR_NONE = 0
    ERROR_TOKEN_INVALID = 1
    ERROR_TOKEN_EXPIRED = 2
    ERROR_INSUFFICIENT_PERMISSIONS = 3
    ERROR_NO_CHARACTER = 4
    ERROR_ESI_UNAVAILABLE = 5
    ERROR_OPERATION_MODE_MISMATCH = 6
    ERROR_UNKNOWN = 99

    ERRORS_LIST = [
        (ERROR_NONE, "No error"),
        (ERROR_TOKEN_INVALID, "Invalid token"),
        (ERROR_TOKEN_EXPIRED, "Expired token"),
        (ERROR_INSUFFICIENT_PERMISSIONS, "Insufficient permissions"),
        (ERROR_NO_CHARACTER, "No character set for fetching data from ESI"),
        (ERROR_ESI_UNAVAILABLE, "ESI API is currently unavailable"),
        (
            ERROR_OPERATION_MODE_MISMATCH,
            "Operaton mode does not match with current setting",
        ),
        (ERROR_UNKNOWN, "Unknown error"),
    ]

    corporation = models.OneToOneField(
        EveCorporationInfo,
        default=None,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        help_text="Corporation owning blueprints, if this is a 'corporate' owner",
        related_name="+",
    )
    character = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text="character used for syncing blueprints",
        related_name="+",
    )
    is_active = models.BooleanField(
        default=True,
        help_text=("whether this owner is currently included in the sync process"),
    )

    class Meta:
        default_permissions = ()

    @property
    def name(self) -> str:
        if self.corporation:
            return self.corporation.corporation_name
        else:
            return self.character.character.character_name

    def update_locations_esi(self):
        if self.corporation:
            assets = self._fetch_corporate_assets()
            token = self.token(
                [
                    "esi-universe.read_structures.v1",
                    "esi-assets.read_corporation_assets.v1",
                ]
            )[0]
        else:
            assets = self._fetch_personal_assets()
            token = self.token(
                ["esi-universe.read_structures.v1", "esi-assets.read_assets.v1"]
            )[0]

        asset_ids = []
        asset_locations = {}
        assets_by_id = {}
        for asset in assets:
            asset_ids.append(asset["item_id"])
            assets_by_id[asset["item_id"]] = asset
        for asset in assets:
            if asset["location_id"] in asset_ids:
                location_id = asset["location_id"]
                if asset_locations.get(location_id):
                    asset_locations[location_id].append(asset["item_id"])
                else:
                    asset_locations[location_id] = [asset["item_id"]]
        for location in list(asset_locations.keys()):
            if not Location.objects.filter(id=location).count() > 0:
                parent_location = assets_by_id[location]["location_id"]
                Location.objects.create(
                    id=location,
                    parent=self._fetch_location(parent_location, token=token),
                    eve_type=EveType.objects.filter(
                        id=assets_by_id[location]["type_id"]
                    ).first(),
                )
            else:
                parent_location = assets_by_id[location]["location_id"]
                location_obj = Location.objects.filter(id=location).first()
                location_obj.parent = self._fetch_location(parent_location, token=token)
                location_obj.save()

    def update_blueprints_esi(self):
        """updates all blueprints from ESI"""

        if self.is_active:
            blueprint_ids_to_remove = list(
                Blueprint.objects.filter(owner=self).values_list("item_id", flat=True)
            )
            if self.corporation:
                blueprints = self._fetch_corporate_blueprints()
                token = self.token(
                    [
                        "esi-universe.read_structures.v1",
                        "esi-corporations.read_blueprints.v1",
                    ]
                )[0]
            else:
                blueprints = self._fetch_personal_blueprints()
                token = self.token(
                    [
                        "esi-universe.read_structures.v1",
                        "esi-characters.read_blueprints.v1",
                    ]
                )[0]

            for blueprint in blueprints:
                runs = blueprint["runs"]
                if runs < 1:
                    runs = None
                quantity = blueprint["quantity"]
                if quantity < 0:
                    quantity = 1
                original = Blueprint.objects.filter(
                    owner=self, item_id=blueprint["item_id"]
                ).first()
                if original is not None:
                    # We've seen this blueprint coming from ESI, so we know it shouldn't be deleted
                    blueprint_ids_to_remove.remove(original.item_id)
                    original.location = self._fetch_location(
                        blueprint["location_id"],
                        token=token,
                    )
                    original.location_flag = blueprint["location_flag"]
                    original.eve_type = EveType.objects.get_or_create_esi(
                        id=blueprint["type_id"]
                    )[0]
                    original.runs = runs
                    original.material_efficiency = blueprint["material_efficiency"]
                    original.time_efficiency = blueprint["time_efficiency"]
                    original.quantity = quantity
                    original.save()
                else:
                    Blueprint.objects.create(
                        owner=self,
                        location=self._fetch_location(
                            blueprint["location_id"],
                            token=token,
                        ),
                        location_flag=blueprint["location_flag"],
                        eve_type=EveType.objects.get_or_create_esi(
                            id=blueprint["type_id"]
                        )[0],
                        item_id=blueprint["item_id"],
                        runs=runs,
                        material_efficiency=blueprint["material_efficiency"],
                        time_efficiency=blueprint["time_efficiency"],
                        quantity=quantity,
                    )
            Blueprint.objects.filter(pk__in=blueprint_ids_to_remove).delete()

    def update_industry_jobs_esi(self):
        """updates all blueprints from ESI"""

        if self.is_active:
            job_ids_to_remove = list(
                IndustryJob.objects.filter(owner=self).values_list("id", flat=True)
            )
            if self.corporation:
                jobs = self._fetch_corporate_industry_jobs()
                token = self.token(
                    [
                        "esi-universe.read_structures.v1",
                        "esi-industry.read_corporation_jobs.v1",
                    ]
                )[0]
            else:
                jobs = self._fetch_personal_industry_jobs()
                token = self.token(
                    [
                        "esi-universe.read_structures.v1",
                        "esi-industry.read_character_jobs.v1",
                    ]
                )[0]

            for job in jobs:

                original = IndustryJob.objects.filter(
                    owner=self, id=job["job_id"]
                ).first()
                blueprint = Blueprint.objects.filter(pk=job["blueprint_id"]).first()
                if blueprint is not None:
                    if original is not None:
                        # We've seen this job coming from ESI, so we know it shouldn't be deleted
                        job_ids_to_remove.remove(original.id)
                        original.status = job["status"]
                        original.save()
                    else:
                        # Reject personal listings of corporate jobs and visa-versa
                        if blueprint.owner == self:
                            installer = EveCharacter.objects.get_character_by_id(
                                job["installer_id"]
                            )
                            if not installer:
                                installer = EveCharacter.objects.create_character(
                                    job["installer_id"]
                                )
                            IndustryJob.objects.create(
                                id=job["job_id"],
                                activity=job["activity_id"],
                                owner=self,
                                location=self._fetch_location(
                                    job["location_id"],
                                    token=token,
                                ),
                                blueprint=Blueprint.objects.get(pk=job["blueprint_id"]),
                                installer=installer,
                                runs=job["runs"],
                                start_date=job["start_date"],
                                end_date=job["end_date"],
                                status=job["status"],
                            )
                else:
                    blueprint_id = job["blueprint_id"]
                    logger.warn(f"Unmatchable blueprint ID: {blueprint_id}")
            IndustryJob.objects.filter(pk__in=job_ids_to_remove).delete()

    @fetch_token_for_owner(["esi-assets.read_corporation_assets.v1"])
    def _fetch_corporate_assets(self, token) -> list:
        return esi.client.Assets.get_corporations_corporation_id_assets(
            corporation_id=self.corporation.corporation_id,
            token=token.valid_access_token(),
        ).results()

    @fetch_token_for_owner(["esi-assets.read_assets.v1"])
    def _fetch_personal_assets(self, token) -> list:
        return esi.client.Assets.get_characters_character_id_assets(
            character_id=self.character.character.character_id,
            token=token.valid_access_token(),
        ).results()

    @fetch_token_for_owner(["esi-corporations.read_blueprints.v1"])
    def _fetch_corporate_blueprints(self, token) -> list:

        corporation_id = self.corporation.corporation_id

        blueprints = esi.client.Corporation.get_corporations_corporation_id_blueprints(
            corporation_id=corporation_id,
            token=token.valid_access_token(),
        ).results()
        return blueprints

    @fetch_token_for_owner(["esi-characters.read_blueprints.v1"])
    def _fetch_personal_blueprints(self, token) -> list:

        character_id = self.character.character.character_id

        blueprints = esi.client.Character.get_characters_character_id_blueprints(
            character_id=character_id,
            token=token.valid_access_token(),
        ).results()
        return blueprints

    @fetch_token_for_owner(["esi-industry.read_corporation_jobs.v1"])
    def _fetch_corporate_industry_jobs(self, token) -> list:

        corporation_id = self.corporation.corporation_id

        jobs = esi.client.Industry.get_corporations_corporation_id_industry_jobs(
            corporation_id=corporation_id,
            token=token.valid_access_token(),
        ).results()
        return jobs

    @fetch_token_for_owner(["esi-industry.read_character_jobs.v1"])
    def _fetch_personal_industry_jobs(self, token) -> list:

        character_id = self.character.character.character_id

        jobs = esi.client.Industry.get_characters_character_id_industry_jobs(
            character_id=character_id,
            token=token.valid_access_token(),
        ).results()
        return jobs

    def token(self, scopes=None) -> Tuple[Token, int]:
        """returns a valid Token for the owner"""
        token = None
        error = None
        add_prefix = self._logger_prefix()

        # abort if character is not configured
        if self.character is None:
            logger.error(add_prefix("No character configured to sync"))
            error = self.ERROR_NO_CHARACTER

        # abort if character does not have sufficient permissions
        elif self.corporation and not self.character.user.has_perm(
            "blueprints.add_corporate_blueprint_owner"
        ):
            logger.error(
                add_prefix(
                    "This character does not have sufficient permission to sync corporations"
                )
            )
            error = self.ERROR_INSUFFICIENT_PERMISSIONS

        # abort if character does not have sufficient permissions
        elif not self.character.user.has_perm(
            "blueprints.add_personal_blueprint_owner"
        ):
            logger.error(
                add_prefix(
                    "This character does not have sufficient permission to sync personal blueprints"
                )
            )
            error = self.ERROR_INSUFFICIENT_PERMISSIONS

        else:
            try:
                # get token
                token = (
                    Token.objects.filter(
                        user=self.character.user,
                        character_id=self.character.character.character_id,
                    )
                    .require_scopes(scopes)
                    .require_valid()
                    .first()
                )
            except TokenInvalidError:
                logger.error(add_prefix("Invalid token for fetching blueprints"))
                error = self.ERROR_TOKEN_INVALID
            except TokenExpiredError:
                logger.error(add_prefix("Token expired for fetching blueprints"))
                error = self.ERROR_TOKEN_EXPIRED
            else:
                if not token:
                    logger.error(add_prefix("No token found with sufficient scopes"))
                    error = self.ERROR_TOKEN_INVALID

        return token, error

    def _fetch_location(self, location_id, token) -> "Location":
        return Location.objects.get_or_create_esi_async(id=location_id, token=token)[0]

    def _logger_prefix(self):
        """returns standard logger prefix function"""
        if self.corporation:
            return make_logger_prefix(self.corporation.corporation_ticker)
        else:
            return make_logger_prefix(self.character.character.character_name)

    def __str__(self):
        if self.corporation:
            return self.corporation.corporation_name
        else:
            return self.character.character.character_name


class Blueprint(models.Model):

    item_id = models.PositiveBigIntegerField(
        primary_key=True, help_text="The EVE Item ID of the blueprint"
    )
    owner = models.ForeignKey(
        Owner,
        on_delete=models.CASCADE,
        help_text="Corporation that owns the blueprint",
    )
    eve_type = models.ForeignKey(
        EveType, on_delete=models.CASCADE, help_text="Blueprint type"
    )
    location = models.ForeignKey(
        "Location", on_delete=models.CASCADE, help_text="Blueprint location"
    )
    _location_flag_choices = []
    for choice in EVE_LOCATION_FLAGS:
        _location_flag_choices.append((choice, choice))

    location_flag = models.CharField(
        help_text="Additional location information",
        choices=_location_flag_choices,
        max_length=36,
    )
    quantity = models.PositiveIntegerField(help_text="Number of blueprints", default=1)
    runs = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text="Runs remaining or null if the blueprint is an original",
    )
    material_efficiency = models.PositiveIntegerField(
        help_text="Material efficiency of the blueprint",
        validators=[validate_material_efficiency],
    )
    time_efficiency = models.PositiveIntegerField(
        help_text="Time efficiency of the blueprint",
        validators=[validate_time_efficiency],
    )

    class Meta:
        default_permissions = ()

    def __str__(self):
        return (
            self.eve_type.name + f" ({self.material_efficiency}/{self.time_efficiency})"
        )

    def has_industryjob(self):
        return hasattr(self, "industryjob") and self.industryjob is not None


class IndustryJob(models.Model):
    id = models.PositiveBigIntegerField(
        primary_key=True,
        help_text=("Eve Online job ID"),
    )

    class Activity(models.IntegerChoices):
        MANUFACTURING = 1, _("Manufacturing")
        RESEARCHING_TECHNOLOGY = 2, _("Researching Technology")
        RESEARCHING_TIME_EFFICIENCY = 3, _("Researching Time Efficiency")
        RESEARCHING_MATERIAL_EFFICIENCY = 4, _("Researching Material Efficiency")
        COPYING = 5, _("Copying")
        DUPLICATING = 6, _("Duplicating")
        REVERSE_ENGINEERING = 7, _("Reverse Engineering")
        INVENTING = 8, _("Inventing")
        REACTING = 9, _("Reacting")

    activity = models.PositiveIntegerField(choices=Activity.choices)
    location = models.ForeignKey(
        "Location",
        on_delete=models.CASCADE,
        help_text=(
            "Eve Online location ID of the facility in which the job is running"
        ),
        related_name="+",
    )
    blueprint = models.OneToOneField(
        Blueprint,
        on_delete=models.CASCADE,
        help_text=("Blueprint the job is running"),
    )
    installer = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE,
        related_name="jobs",
    )
    owner = models.ForeignKey(
        Owner,
        on_delete=models.CASCADE,
        related_name="jobs",
    )
    start_date = models.DateTimeField()
    end_date = models.DateTimeField()
    runs = models.PositiveIntegerField()
    status = models.CharField(
        choices=(
            ("active", "Active"),
            ("cancelled", "Cancelled"),
            ("delivered", "Delivered"),
            ("paused", "Paused"),
            ("ready", "Ready"),
            ("reverted", "Reverted"),
        ),
        max_length=10,
    )


class Location(models.Model):
    """An Eve Online location: Station or Upwell Structure or Solar System"""

    _SOLAR_SYSTEM_ID_START = 30_000_000
    _SOLAR_SYSTEM_ID_END = 33_000_000
    _STATION_ID_START = 60_000_000
    _STATION_ID_END = 64_000_000
    _STRUCTURE_ID_START = 1_000_000_000_000

    id = models.PositiveBigIntegerField(
        primary_key=True,
        help_text=(
            "Eve Online location ID, "
            "either item ID for stations or structure ID for structures"
        ),
    )
    parent = models.ForeignKey(
        "Location",
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text=("Eve Online location ID of the parent object"),
        related_name="+",
    )

    name = models.CharField(
        max_length=NAMES_MAX_LENGTH,
        help_text="In-game name of this station or structure",
    )
    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        related_name="+",
    )
    owner = models.ForeignKey(
        EveEntity,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        blank=True,
        help_text="corporation this station or structure belongs to",
        related_name="+",
    )
    updated_at = models.DateTimeField(auto_now=True)

    objects = LocationManager()

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return self.name_plus + f" [id={self.id}]"

    def __repr__(self) -> str:
        return "{}(id={}, name='{}')".format(
            self.__class__.__name__, self.id, self.name
        )

    @property
    def name_plus(self) -> str:
        """return the actual name or 'Unknown location' for empty locations"""
        if self.is_empty:
            return f"Unknown location #{self.id}"
        if not self.name and self.parent:
            return self.parent.name_plus
        return self.name

    @property
    def is_empty(self) -> bool:
        return not self.eve_solar_system and not self.eve_type and not self.parent_id

    @property
    def solar_system_url(self) -> str:
        """returns dotlan URL for this solar system"""
        try:
            return dotlan.solar_system_url(self.eve_solar_system.name)
        except AttributeError:
            return ""

    @property
    def is_solar_system(self) -> bool:
        return self.is_solar_system_id(self.id)

    @property
    def is_station(self) -> bool:
        return self.is_station_id(self.id)

    @property
    def is_structure(self) -> bool:
        return self.is_structure_id(self.id)

    @classmethod
    def is_solar_system_id(cls, location_id: int) -> bool:
        return cls._SOLAR_SYSTEM_ID_START <= location_id <= cls._SOLAR_SYSTEM_ID_END

    @classmethod
    def is_station_id(cls, location_id: int) -> bool:
        return cls._STATION_ID_START <= location_id <= cls._STATION_ID_END

    @classmethod
    def is_structure_id(cls, location_id: int) -> bool:
        return location_id >= cls._STRUCTURE_ID_START


class Request(models.Model):

    objects = RequestManager()
    blueprint = models.ForeignKey(
        Blueprint,
        on_delete=models.CASCADE,
    )
    requesting_user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="+",
        help_text="The requesting user",
    )
    fulfulling_user = models.ForeignKey(
        User,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="+",
        help_text="The user that fulfilled the request, if it has been fulfilled",
    )
    runs = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text="Runs requested or blank for maximum allowed",
    )
    STATUS_OPEN = "OP"
    STATUS_IN_PROGRESS = "IP"
    STATUS_FULFILLED = "FL"
    STATUS_CANCELLED = "CL"

    STATUS_CHOICES = [
        (STATUS_OPEN, "Open"),
        (STATUS_IN_PROGRESS, "In Progress"),
        (STATUS_FULFILLED, "Fulfilled"),
        (STATUS_CANCELLED, "Cancelled"),
    ]
    status = models.CharField(
        help_text="Status of the blueprint request",
        choices=STATUS_CHOICES,
        max_length=2,
    )

    created_at = models.DateTimeField(auto_now_add=True, db_index=True)

    closed_at = models.DateTimeField(blank=True, null=True, db_index=True)

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.requesting_user.profile.main_character.character_name}'s request for {self.blueprint.eve_type.name}"

    def __repr__(self) -> str:
        return "{}(id={}, requesting_user='{}', type_name='{}')".format(
            self.__class__.__name__,
            self.pk,
            self.requesting_user.profile.main_character.character_name,
            self.eve_type.name,
        )



from esi.clients import EsiClientProvider

from . import __version__
from .utils import get_swagger_spec_path

esi = EsiClientProvider(
    spec_file=get_swagger_spec_path(), app_info_text=f"aa-blueprints v{__version__}"
)



from django.utils.translation import ugettext_lazy as _

from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

from . import urls
from .models import Request


class BlueprintLibraryMenuItem(MenuItemHook):
    """ This class ensures only authorized users will see the menu entry """

    def __init__(self):
        # setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            _("Blueprint Library"),
            "fas fa-scroll fa-fw",
            "blueprints:index",
            navactive=["blueprints:index"],
        )

    def render(self, request):
        if request.user.has_perm("blueprints.basic_access"):
            if request.user.has_perm("blueprints.manage_requests"):
                app_count = (
                    Request.objects.all()
                    .requests_fulfillable_by_user(request.user)
                    .count()
                    + Request.objects.all()
                    .requests_being_fulfilled_by_user(request.user)
                    .count()
                )
            else:
                app_count = None
            self.count = app_count if app_count and app_count > 0 else None
            return MenuItemHook.render(self, request)
        return ""


@hooks.register("menu_item_hook")
def register_menu():
    return BlueprintLibraryMenuItem()


@hooks.register("url_hook")
def register_urls():
    return UrlHook(urls, "blueprints", r"^blueprints/")



from datetime import datetime

from django.contrib.auth.decorators import login_required
from django.db import transaction
from django.db.models import Q
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.html import format_html
from django.utils.translation import gettext_lazy
from django.views.decorators.http import require_POST

from allianceauth.authentication.decorators import permissions_required
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from esi.decorators import token_required
from eveuniverse.models import EveType

from . import __title__, tasks
from .app_settings import (
    BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED,
    BLUEPRINTS_DEFAULT_PAGE_LENGTH,
    BLUEPRINTS_LIST_ICON_OUTPUT_SIZE,
    BLUEPRINTS_PAGING_ENABLED,
)
from .models import Blueprint, Owner, Request
from .utils import messages_plus, notify_admins


@login_required
@permissions_required("blueprints.basic_access")
def index(request):
    if request.user.has_perm("blueprints.manage_requests"):
        request_count = (
            Request.objects.all().requests_fulfillable_by_user(request.user).count()
            + Request.objects.all()
            .requests_being_fulfilled_by_user(request.user)
            .count()
        )
    else:
        request_count = None
    context = {
        "page_title": gettext_lazy(__title__),
        "data_tables_page_length": BLUEPRINTS_DEFAULT_PAGE_LENGTH,
        "data_tables_paging": BLUEPRINTS_PAGING_ENABLED,
        "request_count": request_count,
    }
    return render(request, "blueprints/index.html", context)


@login_required
@permissions_required("blueprints.add_corporate_blueprint_owner")
@token_required(
    scopes=[
        "esi-universe.read_structures.v1",
        "esi-corporations.read_blueprints.v1",
        "esi-assets.read_corporation_assets.v1",
        "esi-industry.read_corporation_jobs.v1",
    ]
)
def add_corporate_blueprint_owner(request, token):
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    success = True
    try:
        owned_char = CharacterOwnership.objects.get(
            user=request.user, character=token_char
        )
    except CharacterOwnership.DoesNotExist:
        messages_plus.error(
            request,
            format_html(
                gettext_lazy(
                    "You can only use your main or alt characters "
                    "to add corporations. "
                    "However, character %s is neither. "
                )
                % format_html("<strong>{}</strong>", token_char.character_name)
            ),
        )
        success = False
        owned_char = None

    if success:
        try:
            corporation = EveCorporationInfo.objects.get(
                corporation_id=token_char.corporation_id
            )
        except EveCorporationInfo.DoesNotExist:
            corporation = EveCorporationInfo.objects.create_corporation(
                token_char.corporation_id
            )

        with transaction.atomic():
            owner, _ = Owner.objects.update_or_create(
                corporation=corporation, defaults={"character": owned_char}
            )

            owner.save()

        tasks.update_blueprints_for_owner.delay(owner_pk=owner.pk)
        tasks.update_locations_for_owner.delay(owner_pk=owner.pk)
        messages_plus.info(
            request,
            format_html(
                gettext_lazy(
                    "%(corporation)s has been added with %(character)s "
                    "as sync character. We have started fetching blueprints "
                    "for this corporation. You will receive a report once "
                    "the process is finished."
                )
                % {
                    "corporation": format_html("<strong>{}</strong>", owner),
                    "character": format_html(
                        "<strong>{}</strong>", owner.character.character.character_name
                    ),
                }
            ),
        )
        if BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED:
            notify_admins(
                message=gettext_lazy(
                    "%(corporation)s was added as new corporate blueprint owner by %(user)s."
                )
                % {
                    "corporation": owner.corporation.corporation_name,
                    "user": request.user.username,
                },
                title="{}: blueprint owner added: {}".format(
                    __title__, owner.corporation.corporation_name
                ),
            )
    return redirect("blueprints:index")


@login_required
@permissions_required("blueprints.add_personal_blueprint_owner")
@token_required(
    scopes=[
        "esi-universe.read_structures.v1",
        "esi-characters.read_blueprints.v1",
        "esi-assets.read_assets.v1",
        "esi-industry.read_character_jobs.v1",
    ]
)
def add_personal_blueprint_owner(request, token):
    token_char = EveCharacter.objects.get(character_id=token.character_id)

    success = True
    try:
        owned_char = CharacterOwnership.objects.get(
            user=request.user, character=token_char
        )
    except CharacterOwnership.DoesNotExist:
        messages_plus.error(
            request,
            format_html(
                gettext_lazy(
                    "You can only use your main or alt characters "
                    "to add corporations. "
                    "However, character %s is neither. "
                )
                % format_html("<strong>{}</strong>", token_char.character_name)
            ),
        )
        success = False
        owned_char = None

    if success:

        with transaction.atomic():
            owner, _ = Owner.objects.update_or_create(
                corporation=None, character=owned_char
            )

            owner.save()

        tasks.update_blueprints_for_owner.delay(owner_pk=owner.pk)
        tasks.update_locations_for_owner.delay(owner_pk=owner.pk)
        messages_plus.info(
            request,
            format_html(
                gettext_lazy(
                    "%(character)s has been added. We have started fetching blueprints "
                    "for this character. You will receive a report once "
                    "the process is finished."
                )
                % {
                    "character": format_html(
                        "<strong>{}</strong>", owner.character.character.character_name
                    ),
                }
            ),
        )
        if BLUEPRINTS_ADMIN_NOTIFICATIONS_ENABLED:
            notify_admins(
                message=gettext_lazy(
                    "%(character)s was added as a new personal blueprint owner."
                )
                % {
                    "character": owner.character.character.character_name,
                },
                title="{}: blueprint owner added: {}".format(
                    __title__, owner.character.character.character_name
                ),
            )
    return redirect("blueprints:index")


def convert_blueprint(blueprint: Blueprint, user, details=False) -> dict:
    variant = EveType.IconVariant.BPC if blueprint.runs else EveType.IconVariant.BPO
    icon = format_html(
        '<img src="{}" width="{}" height="{}">',
        blueprint.eve_type.icon_url(size=64, variant=variant),
        BLUEPRINTS_LIST_ICON_OUTPUT_SIZE,
        BLUEPRINTS_LIST_ICON_OUTPUT_SIZE,
    )
    runs = "" if not blueprint.runs or blueprint.runs == -1 else blueprint.runs
    original = "✓" if not blueprint.runs or blueprint.runs == -1 else ""
    filter_is_original = (
        gettext_lazy("Yes")
        if not blueprint.runs or blueprint.runs == -1
        else gettext_lazy("No")
    )
    if blueprint.owner.corporation:
        owner_type = "corporation"
    else:
        owner_type = "character"

    summary = {
        "icn": icon,
        "qty": blueprint.quantity,
        "pk": blueprint.pk,
        "nme": blueprint.eve_type.name,
        "loc": blueprint.location.name_plus,
        "me": blueprint.material_efficiency,
        "te": blueprint.time_efficiency,
        "og": original,
        "iog": filter_is_original,
        "rns": runs,
        "on": blueprint.owner.name,
        "ot": owner_type,
        "use": blueprint.has_industryjob(),
    }
    if details:
        if blueprint.has_industryjob() and user.has_perm(
            "blueprints.view_industry_jobs"
        ):
            job = {
                "activity": blueprint.industryjob.get_activity_display(),
                "installer": blueprint.industryjob.installer.character_name,
                "runs": blueprint.industryjob.runs,
                "start_date": blueprint.industryjob.start_date,
                "end_date": blueprint.industryjob.end_date,
            }
            summary.update({"job": job})
        summary.update({"frm": blueprint.eve_type.name.endswith(" Formula")})
    return summary


@login_required
@permissions_required("blueprints.basic_access")
def list_blueprints(request):

    corporation_ids = set(
        request.user.character_ownerships.select_related("character").values_list(
            "character__corporation_id", flat=True
        )
    )
    corporations = list(
        EveCorporationInfo.objects.filter(corporation_id__in=corporation_ids)
    )
    if request.user.has_perm("blueprints.view_alliance_blueprints"):
        alliances = {
            corporation.alliance for corporation in corporations if corporation.alliance
        }
        for alliance in alliances:
            corporations += alliance.evecorporationinfo_set.all()

        corporations = list(set(corporations))

    personal_owner_ids = list()
    for owner in Owner.objects.filter(corporation=None):
        if owner.character.character.corporation_id in corporation_ids:
            personal_owner_ids.append(owner.pk)

    blueprints_query = Blueprint.objects.filter(
        Q(owner__corporation__in=corporations) | Q(owner__pk__in=personal_owner_ids)
    ).select_related(
        "eve_type",
        "location",
        "industryjob",
        "owner",
        "owner__corporation",
        "owner__character",
        "location",
        "location__eve_solar_system",
        "location__eve_type",
    )
    blueprint_rows = [
        convert_blueprint(blueprint, request.user) for blueprint in blueprints_query
    ]

    return JsonResponse(blueprint_rows, safe=False)


@login_required
@permissions_required(
    "blueprints.add_personal_blueprint_owner",
    "blueprints.add_corporate_blueprint_owner",
)
def list_user_owners(request):
    owners = Owner.objects.filter(character__user=request.user)
    results = []
    for owner in owners:
        if owner.corporation:
            owner_type = "corporate"
            owner_type_display = gettext_lazy("Corporate")
            owner_name = owner.corporation.corporation_name
        else:
            owner_type = "personal"
            owner_type_display = gettext_lazy("Personal")
            owner_name = owner.character.character.character_name
        results.append(
            {
                "id": owner.pk,
                "type": owner_type,
                "type_display": owner_type_display,
                "name": owner_name,
            }
        )
    return JsonResponse(results, safe=False)


@login_required
def view_blueprint_modal(request):
    blueprint = Blueprint.objects.get(pk=request.GET.get("blueprint_id"))
    context = {"blueprint": convert_blueprint(blueprint, request.user, details=True)}
    return render(request, "blueprints/modals/view_blueprint_content.html", context)


@login_required
@permissions_required(("blueprints.request_blueprints", "blueprints.manage_requests"))
def view_request_modal(request):
    user_request = Request.objects.get(pk=request.GET.get("request_id"))
    context = {"request": convert_request(user_request)}
    return render(request, "blueprints/modals/view_request_content.html", context)


@login_required
@permissions_required("blueprints.request_blueprints")
def create_request(request):
    if request.method == "POST":
        requested = Blueprint.objects.get(pk=request.POST.get("pk"))
        runs = request.POST.get("runs")
        if runs == "":
            runs = None
        user = request.user
        Request.objects.create(
            blueprint=requested,
            requesting_user=user,
            status=Request.STATUS_OPEN,
            runs=runs,
        )
        messages_plus.info(
            request,
            format_html(
                gettext_lazy("A copy of %(blueprint)s has been requested.")
                % {"blueprint": requested.eve_type.name}
            ),
        )
    return redirect("blueprints:index")


def convert_request(request: Request) -> dict:
    variant = (
        EveType.IconVariant.BPC if request.blueprint.runs else EveType.IconVariant.BPO
    )
    icon = format_html(
        '<img src="{}" width="{}" height="{}">',
        request.blueprint.eve_type.icon_url(size=64, variant=variant),
        BLUEPRINTS_LIST_ICON_OUTPUT_SIZE,
        BLUEPRINTS_LIST_ICON_OUTPUT_SIZE,
    )

    if request.blueprint.owner.corporation:
        owner_type = "corporation"
    else:
        owner_type = "character"
    return {
        "request_id": request.pk,
        "type_icon": icon,
        "type_name": request.blueprint.eve_type.name,
        "owner_name": request.blueprint.owner.name,
        "owner_type": owner_type,
        "requestor": request.requesting_user.profile.main_character.character_name,
        "location": request.blueprint.location.name_plus,
        "material_efficiency": request.blueprint.material_efficiency,
        "time_efficiency": request.blueprint.time_efficiency,
        "runs": request.runs if request.runs else "",
        "status": request.status,
        "status_display": request.get_status_display(),
    }


@login_required
@permissions_required("blueprints.request_blueprints")
def list_user_requests(request):

    request_rows = list()

    request_query = Request.objects.select_related_default().filter(
        requesting_user=request.user, closed_at=None
    )
    for request in request_query:
        request_rows.append(convert_request(request))

    return JsonResponse(request_rows, safe=False)


@login_required
@permissions_required("blueprints.manage_requests")
def list_open_requests(request):

    request_rows = list()

    requests = Request.objects.select_related_default().requests_fulfillable_by_user(
        request.user
    ) | Request.objects.select_related_default().requests_being_fulfilled_by_user(
        request.user
    )

    for request in requests:
        request_rows.append(convert_request(request))

    return JsonResponse(request_rows, safe=False)


def mark_request(
    request, request_id, status, fulfilling_user, closed, *, can_requestor_edit=False
):
    completed = False
    user_request = Request.objects.get(pk=request_id)

    corporation_ids = {
        character.character.corporation_id
        for character in request.user.character_ownerships.all()
    }
    character_ownership_ids = {
        character.pk for character in request.user.character_ownerships.all()
    }
    if (
        (
            user_request.blueprint.owner.corporation
            and user_request.blueprint.owner.corporation.corporation_id
            in corporation_ids
        )
        or (
            not user_request.blueprint.owner.corporation
            and user_request.blueprint.owner.pk in character_ownership_ids
        )
        or (can_requestor_edit and user_request.requesting_user == request.user)
    ):
        if closed:
            user_request.closed_at = datetime.utcnow()
        else:
            user_request.closed_at = None
        user_request.fulfulling_user = fulfilling_user
        user_request.status = status
        user_request.save()
        completed = True
    return user_request, completed


@login_required
@permissions_required("blueprints.manage_requests")
@require_POST
def mark_request_fulfilled(request, request_id):
    user_request, completed = mark_request(
        request, request_id, Request.STATUS_FULFILLED, request.user, True
    )
    if completed:
        messages_plus.info(
            request,
            format_html(
                gettext_lazy(
                    "The request for %(blueprint)s has been closed as fulfilled."
                )
                % {"blueprint": user_request.blueprint.eve_type.name}
            ),
        )
    else:
        messages_plus.error(
            request,
            format_html(
                gettext_lazy("Fulfilling the request for %(blueprint)s has failed.")
                % {"blueprint": user_request.blueprint.eve_type.name}
            ),
        )
    return redirect("blueprints:index")


@login_required
@permissions_required("blueprints.manage_requests")
@require_POST
def mark_request_in_progress(request, request_id):
    user_request, completed = mark_request(
        request, request_id, Request.STATUS_IN_PROGRESS, request.user, False
    )
    if completed:
        messages_plus.info(
            request,
            format_html(
                gettext_lazy(
                    "The request for %(blueprint)s has been marked as in progress."
                )
                % {"blueprint": user_request.blueprint.eve_type.name}
            ),
        )
    else:
        messages_plus.error(
            request,
            format_html(
                gettext_lazy(
                    "Marking the request for %(blueprint)s as in progress has failed."
                )
                % {"blueprint": user_request.blueprint.eve_type.name}
            ),
        )
    return redirect("blueprints:index")


@login_required
@permissions_required("blueprints.manage_requests")
@require_POST
def mark_request_open(request, request_id):
    user_request, completed = mark_request(
        request, request_id, Request.STATUS_OPEN, None, False
    )
    if completed:
        messages_plus.info(
            request,
            format_html(
                gettext_lazy("The request for %(blueprint)s has been re-opened.")
                % {"blueprint": user_request.blueprint.eve_type.name}
            ),
        )
    else:
        messages_plus.error(
            request,
            format_html(
                gettext_lazy("Re-opening the request for %(blueprint)s has failed.")
                % {"blueprint": user_request.blueprint.eve_type.name}
            ),
        )
    return redirect("blueprints:index")


@login_required
@permissions_required(["blueprints.basic_access", "blueprints.manage_requests"])
@require_POST
def mark_request_cancelled(request, request_id):
    user_request, completed = mark_request(
        request,
        request_id,
        Request.STATUS_CANCELLED,
        None,
        True,
        can_requestor_edit=True,
    )
    if completed:
        messages_plus.info(
            request,
            format_html(
                gettext_lazy(
                    "The request for %(blueprint)s has been closed as cancelled."
                )
                % {"blueprint": user_request.blueprint.eve_type.name}
            ),
        )
    else:
        messages_plus.error(
            request,
            format_html(
                gettext_lazy("Cancelling the request for %(blueprint)s has failed.")
                % {"blueprint": user_request.blueprint.eve_type.name}
            ),
        )
    return redirect("blueprints:index")


@login_required
@permissions_required(
    "blueprints.add_personal_blueprint_owner",
    "blueprints.add_corporate_blueprint_owner",
)
@require_POST
def remove_owner(request, owner_id):
    owner = Owner.objects.filter(pk=owner_id, character__user=request.user).first()
    completed = False
    owner_name = None

    if owner:
        owner_name = (
            owner.corporation.corporation_name
            if owner.corporation
            else owner.character.character.character_name
        )
        owner.delete()
        completed = True

    if completed:
        messages_plus.info(
            request,
            format_html(
                gettext_lazy("%(owner)s has been removed as a blueprint owner.")
                % {"owner": owner_name}
            ),
        )
    else:
        messages_plus.error(
            request,
            format_html(gettext_lazy("Removing the blueprint owner has failed.")),
        )
    return redirect("blueprints:index")



from django.contrib import admin

from .models import Blueprint, IndustryJob, Location, Owner, Request

# Register your models here.


@admin.register(Blueprint)
class BlueprintAdmin(admin.ModelAdmin):
    list_display = (
        "_type",
        "_owner",
        "material_efficiency",
        "time_efficiency",
        "_original",
    )

    list_select_related = ("eve_type", "owner")
    search_fields = ["eve_type__name"]

    def _type(self, obj):
        return obj.eve_type.name if obj.eve_type else None

    def _owner(self, obj):
        return obj.owner.name

    def _original(self, obj):
        return "No" if obj.runs and obj.runs > 0 else "Yes"

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Location)
class LocationAdmin(admin.ModelAdmin):
    list_display = ("id", "_name", "_type", "_group", "_solar_system", "updated_at")
    list_filter = (
        (
            "eve_solar_system__eve_constellation__eve_region",
            admin.RelatedOnlyFieldListFilter,
        ),
        ("eve_solar_system", admin.RelatedOnlyFieldListFilter),
        ("eve_type__eve_group", admin.RelatedOnlyFieldListFilter),
    )
    search_fields = ["name"]
    list_select_related = (
        "eve_solar_system",
        "eve_solar_system__eve_constellation__eve_region",
        "eve_type",
        "eve_type__eve_group",
    )

    def _name(self, obj):
        if obj.name:
            return obj.name
        if obj.parent and obj.parent.name:
            return "Child of " + obj.parent.name
        return None

    _name.admin_order_field = "name"

    def _solar_system(self, obj):
        return obj.eve_solar_system.name if obj.eve_solar_system else None

    _solar_system.admin_order_field = "eve_solar_system__name"

    def _type(self, obj):
        return obj.eve_type.name if obj.eve_type else None

    _type.admin_order_field = "eve_type__name"

    def _group(self, obj):
        return obj.eve_type.eve_group.name if obj.eve_type else None

    _group.admin_order_field = "eve_type__eve_group__name"

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Owner)
class OwnerAdmin(admin.ModelAdmin):
    list_display = ("character", "_type", "corporation", "is_active")

    def _type(self, obj):
        return "Corporate" if obj.corporation else "Personal"

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False


@admin.register(Request)
class RequestAdmin(admin.ModelAdmin):
    list_display = ("_type", "_requestor", "_owner", "_fulfilled_by")

    list_select_related = (
        "blueprint__eve_type",
        "requesting_user__profile__main_character__character_name",
    )
    search_fields = ["blueprint__eve_type__name"]

    def _type(self, obj):
        return obj.blueprint.eve_type.name if obj.blueprint.eve_type else None

    def _requestor(self, obj):
        return obj.requesting_user.profile.main_character.character_name

    def _owner(self, obj):
        return obj.blueprint.owner.name

    def _fulfilled_by(self, obj):
        return (
            obj.fulfulling_user.profile.main_character.character_name
            if obj.fulfulling_user
            else None
        )

    def has_add_permission(self, request):
        return False


@admin.register(IndustryJob)
class IndustryJobAdmin(admin.ModelAdmin):

    list_display = ("_blueprint", "_installer", "_activity")

    list_select_related = ("blueprint__eve_type",)
    search_fields = ["blueprint__eve_type__name"]

    def _blueprint(self, obj):
        return obj.blueprint.eve_type.name if obj.blueprint.eve_type else None

    def _installer(self, obj):
        return obj.installer.character_name

    def _activity(self, obj):
        return obj.get_activity_display()

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False



import logging

from django.core.management import call_command
from django.core.management.base import BaseCommand

from ... import __title__
from ...constants import EVE_CATEGORY_ID_BLUEPRINT
from ...utils import LoggerAddTag

logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class Command(BaseCommand):
    help = "Preloads data required for aa-blueprints from ESI"

    def handle(self, *args, **options):
        call_command(
            "eveuniverse_load_types",
            __title__,
            "--category_id",
            str(EVE_CATEGORY_ID_BLUEPRINT),
        )
